use super::{Device, Queue, VkAllocator};
use ash::{prelude::*, vk};
use vk_mem::Alloc;

pub struct Image2D {
    alloc: VkAllocator,
    image: vk::Image,
    size: vk::Extent2D,
    mip_levels: u32,
    format: vk::Format,
    allocation: vk_mem::Allocation,
}

impl Image2D {
    pub fn new_exclusive(
        alloc: VkAllocator,
        extent: vk::Extent2D,
        mip_levels: u32,
        array_layers: u32,
        format: vk::Format,
        tiling: vk::ImageTiling,
        usage: vk::ImageUsageFlags,
        memory_usage: vk_mem::MemoryUsage,
        alloc_flags: vk_mem::AllocationCreateFlags,
    ) -> VkResult<Image2D> {
        let create_info = vk::ImageCreateInfo::default()
            .array_layers(array_layers)
            .mip_levels(mip_levels)
            .extent(vk::Extent3D {
                width: extent.width,
                height: extent.height,
                depth: 1,
            })
            .image_type(vk::ImageType::TYPE_2D)
            .format(format)
            .tiling(tiling)
            .initial_layout(vk::ImageLayout::UNDEFINED)
            .usage(usage)
            .sharing_mode(vk::SharingMode::EXCLUSIVE)
            .samples(vk::SampleCountFlags::TYPE_1);

        let alloc_info = vk_mem::AllocationCreateInfo {
            usage: memory_usage,
            flags: alloc_flags,
            ..Default::default()
        };

        let (image, allocation) =
            unsafe { alloc.create_image(&create_info, &alloc_info)? };

        Ok(Self {
            alloc,
            image,
            size: extent,
            mip_levels,
            format,
            allocation,
        })
    }

    pub fn view(
        &self,
        device: &Device,
        aspect: vk::ImageAspectFlags,
    ) -> VkResult<vk::ImageView> {
        let create_info = vk::ImageViewCreateInfo::default()
            .image(self.image)
            .view_type(vk::ImageViewType::TYPE_2D)
            .format(self.format)
            .components(vk::ComponentMapping::default())
            .subresource_range(
                vk::ImageSubresourceRange::default()
                    .aspect_mask(aspect)
                    .base_mip_level(0)
                    .level_count(self.mip_levels)
                    .base_array_layer(0)
                    .layer_count(1),
            );

        let view =
            unsafe { device.dev().create_image_view(&create_info, None)? };

        Ok(view)
    }

    pub fn image(&self) -> vk::Image {
        self.image
    }
}

impl Drop for Image2D {
    fn drop(&mut self) {
        unsafe {
            self.alloc.destroy_image(self.image, &mut self.allocation);
        }
    }
}

pub struct QueueOwnership {
    src: u32,
    dst: u32,
}

pub fn image_barrier<'a>(
    image: vk::Image,
    aspects: vk::ImageAspectFlags,
    src_stage: vk::PipelineStageFlags2,
    src_access: vk::AccessFlags2,
    dst_stage: vk::PipelineStageFlags2,
    dst_access: vk::AccessFlags2,
    old_layout: vk::ImageLayout,
    new_layout: vk::ImageLayout,
    queue_ownership_op: Option<QueueOwnership>,
) -> vk::ImageMemoryBarrier2<'a> {
    let (src_family, dst_family) = queue_ownership_op
        .map(|t| (t.src, t.dst))
        .unwrap_or((vk::QUEUE_FAMILY_IGNORED, vk::QUEUE_FAMILY_IGNORED));

    vk::ImageMemoryBarrier2::default()
        .image(image)
        .subresource_range(
            vk::ImageSubresourceRange::default()
                .aspect_mask(aspects)
                .base_mip_level(0)
                .base_array_layer(0)
                .level_count(vk::REMAINING_MIP_LEVELS)
                .layer_count(vk::REMAINING_ARRAY_LAYERS),
        )
        .src_stage_mask(src_stage)
        .src_access_mask(src_access)
        .dst_stage_mask(dst_stage)
        .dst_access_mask(dst_access)
        .dst_queue_family_index(dst_family)
        .src_queue_family_index(src_family)
        .old_layout(old_layout)
        .new_layout(new_layout)
}
